@@ -1,7 +1,10 @@
 from __future__ import unicode_literals
+import sys
+import warnings
+
 
 __title__ = 'cpymad'
-__version__ = '1.8.0'
+__version__ = '1.9.1'
 
 __summary__ = 'Cython binding to MAD-X'
 __uri__ = 'https://github.com/hibtc/cpymad'
@@ -18,6 +21,14 @@ Initial pymad creators:
 """
 
 
-def get_copyright_notice():
+if sys.version_info < (3, 6):
+    _unsupported_version = (
+        "Support for python 3.5 and below will be removed in a future release!\n"
+        "If you need continued support for an older version, let us know at:\n"
+        "  https://github.com/hibtc/cpymad/issues")
+    warnings.warn(_unsupported_version, DeprecationWarning)
+
+
+def get_copyright_notice() -> str:
     from importlib_resources import read_text
     return read_text('cpymad.COPYING', 'cpymad.rst', encoding='utf-8')
