@@ -14,7 +14,6 @@ import collections.abc as abc
 import os
 import subprocess
 import sys
-import warnings
 
 import numpy as np
 
@@ -23,21 +22,32 @@ from . import util
 from .stream import AsyncReader
 
 
-if sys.version_info < (3, 6):
-    _unsupported_version = (
-        "Support for python 3.5 and below will be removed in a future release!\n"
-        "If you need continued support for an older version, let us know at:\n"
-        "  https://github.com/hibtc/cpymad/issues")
-    warnings.warn(_unsupported_version, DeprecationWarning)
-
-
 __all__ = [
     'Madx',
-    'Sequence',
-    'BaseElementList',
-    'Table',
+    'ArrayAttribute',
+    'AttrDict',
+    'BaseTypeMap',
+    'Command',
     'CommandLog',
+    'CommandMap',
+    'Element',
+    'ElementList',
+    'ExpandedElementList',
+    'GlobalElementList',
+    'Metadata',
+    'Sequence',
+    'SequenceMap',
+    'Table',
+    'TableMap',
+    'VarList',
+    'VarParamList',
+    'Version',
+
+    # Data
     'metadata',
+
+    # Errors:
+    'TwissFailed',
 ]
 
 
@@ -45,7 +55,7 @@ class TwissFailed(RuntimeError):
     pass
 
 
-class Version(object):
+class Version:
 
     """Version information struct. """
 
@@ -60,11 +70,11 @@ class Version(object):
         return "MAD-X {} ({})".format(self.release, self.date)
 
 
-class NullContext(object):
+class NullContext:
     __enter__ = __exit__ = lambda *_: None
 
 
-class CommandLog(object):
+class CommandLog:
 
     """Log MAD-X command history to a text file."""
 
@@ -83,7 +93,7 @@ class CommandLog(object):
     def __del__(self):
         self.close()
 
-    def __call__(self, command):
+    def __call__(self, command: str):
         """Log a single history line and flush to file immediately."""
         self._file.write(self._prefix + command + self._suffix)
         self._file.flush()
@@ -93,7 +103,7 @@ class CommandLog(object):
             self._file.close()
 
 
-class Madx(object):
+class Madx:
 
     """
     Python interface for a MAD-X process.
@@ -122,13 +132,14 @@ class Madx(object):
     """
 
     def __init__(self, libmadx=None, command_log=None, stdout=None,
-                 history=None, **Popen_args):
+                 history=None, prompt=None, **Popen_args):
         """
         Initialize instance variables.
 
         :param libmadx: :mod:`libmadx` compatible object
         :param command_log: Log all MAD-X commands issued via cpymad.
         :param stdout: file descriptor, file object or callable
+        :param str prompt: prefix for a new :class:`CommandLog`
         :param Popen_args: Additional parameters to ``subprocess.Popen``
 
         If ``libmadx`` is NOT specified, a new MAD-X interpreter will
@@ -144,9 +155,17 @@ class Madx(object):
 
             m = Madx(stdout=sys.stdout)
         """
-        # open history file
         if isinstance(command_log, str):
-            command_log = CommandLog.create(command_log)
+            # open new history file:
+            command_log = CommandLog.create(command_log, prompt or '')
+        elif hasattr(command_log, 'write'):
+            # assuming stream already opened:
+            command_log = CommandLog(command_log, prompt or '')
+        elif prompt is not None:
+            assert command_log is None, \
+                "Passing fully constructed `command_log` instances is " \
+                "incompatible with parameter `prompt`."
+            command_log = CommandLog(sys.stdout, prompt)
         self.reader = NullContext()
         # start libmadx subprocess
         if libmadx is None:
@@ -198,9 +217,8 @@ class Madx(object):
         try:
             return getattr(self.command, name)
         except AttributeError:
-            pass
-        raise AttributeError('Unknown attribute or command: {!r}'
-                             .format(name))
+            raise AttributeError(
+                'Unknown attribute or command: {!r}'.format(name)) from None
 
     def quit(self):
         """Shutdown MAD-X interpreter and stop process."""
@@ -226,25 +244,24 @@ class Madx(object):
     # Data descriptors:
 
     @property
-    def version(self):
+    def version(self) -> Version:
         """Get the MAD-X version."""
         return Version(self._libmadx.get_version_number(),
                        self._libmadx.get_version_date())
 
     @property
-    def options(self):
+    def options(self) -> "Command":
         """Values of current options."""
         return Command(self, self._libmadx.get_options())
 
     # Methods:
 
-    def input(self, text):
+    def input(self, text: str) -> bool:
         """
         Run any textual MAD-X input.
 
-        :param str text: command text
+        :param text: command text
         :returns: whether the command has completed without error
-        :rtype: bool
         """
         text = text.rstrip(';') + ';'
         if self._enter_count > 0:
@@ -260,9 +277,7 @@ class Madx(object):
             with self.reader:
                 return self._libmadx.input(text)
         except _rpc.RemoteProcessCrashed:
-            # catch + reraise in order to shorten stack trace (~3-5 levels):
-            pass
-        raise RuntimeError("MAD-X has stopped working!")
+            raise RuntimeError("MAD-X has stopped working!") from None
 
     __call__ = input
 
@@ -289,7 +304,7 @@ class Madx(object):
                 self.input("\n".join(self._batch))
                 self._batch = None
 
-    def expr_vars(self, expr):
+    def expr_vars(self, expr: str) -> list:
         """Find all variable names used in an expression. This does *not*
         include element attribute nor function names."""
         if not isinstance(expr, str):
@@ -299,13 +314,12 @@ class Madx(object):
                 and v in self.globals
                 and self._libmadx.get_var_type(v) > 0]
 
-    def chdir(self, dir):
+    def chdir(self, dir: str) -> util.ChangeDirectory:
         """
         Change the directory of the MAD-X process (not the current python process).
 
-        :param str path: new path name
+        :param dir: new path name
         :returns: a context manager that can change the directory back
-        :rtype: ChangeDirectory
 
         It can be used as context manager for temporary directory changes::
 
@@ -316,15 +330,15 @@ class Madx(object):
         return util.ChangeDirectory(dir, self._chdir, self._libmadx.getcwd)
 
     # Turns `dir` into a keyword argument for CHDIR command:
-    def _chdir(self, dir):
+    def _chdir(self, dir: str):
         return self.command.chdir(dir=dir)
 
-    def call(self, file, chdir=False):
+    def call(self, file: str, chdir: bool = False):
         """
         CALL a file in the MAD-X interpreter.
 
-        :param str file: file name with path
-        :param bool chdir: temporarily change directory in MAD-X process
+        :param file: file name with path
+        :param chdir: temporarily change directory in MAD-X process
         """
         if chdir:
             dirname, basename = os.path.split(file)
@@ -362,7 +376,7 @@ class Madx(object):
             self._libmadx.apply_table_selections(table)
         return self.table[table]
 
-    def use(self, sequence=None, range=None, **kwargs):
+    def use(self, sequence: str = None, range: str = None, **kwargs):
         """
         Run USE to expand a sequence.
 
@@ -406,7 +420,7 @@ class Madx(object):
               method=('lmdif', {}),
               knobfile=None,
               limits=None,
-              **kwargs):
+              **kwargs) -> dict:
         """
         Perform a simple MATCH operation.
 
@@ -418,7 +432,6 @@ class Madx(object):
         :param str knobfile: file to write the knob values to
         :param dict kwargs: keyword arguments for the MAD-X command
         :returns: final knob values
-        :rtype: dict
 
         Example:
 
@@ -458,21 +471,17 @@ class Madx(object):
         """Turn verbose output on/off."""
         self.command.option(echo=switch, warn=switch, info=switch)
 
-    def eval(self, expr):
+    def eval(self, expr) -> float:
         """
         Evaluates an expression and returns the result as double.
 
         :param str expr: expression to evaluate.
         :returns: numeric value of the expression
-        :rtype: float
         """
         if isinstance(expr, (float, int, bool)):
             return expr
         if isinstance(expr, (list, ArrayAttribute)):
             return [self.eval(x) for x in expr]
-        # Try to prevent process crashes:
-        # NOTE: this limits to a sane subset of accepted MAD-X expressions.
-        util.check_expression(expr)
         return self._libmadx.eval(expr)
 
 
@@ -490,8 +499,7 @@ class _Mapping(abc.Mapping):
         try:
             return self[key]
         except KeyError:
-            pass
-        raise AttributeError(self._missing(key))
+            raise AttributeError(self._missing(key)) from None
 
     def _missing(self, key):
         return key
@@ -554,8 +562,7 @@ class SequenceMap(_Mapping):
         try:
             return Sequence(name, self._madx)
         except ValueError:
-            pass
-        raise KeyError("Unknown sequence: {!r}".format(name))
+            raise KeyError("Unknown sequence: {!r}".format(name)) from None
 
     def __contains__(self, name):
         return self._libmadx.sequence_exists(name.lower())
@@ -586,8 +593,7 @@ class TableMap(_Mapping):
         try:
             return Table(name, self._libmadx)
         except ValueError:
-            pass
-        raise KeyError("Table not found {!r}".format(name))
+            raise KeyError("Table not found {!r}".format(name)) from None
 
     def __contains__(self, name):
         return self._libmadx.table_exists(name.lower())
@@ -596,7 +602,7 @@ class TableMap(_Mapping):
         return self._libmadx.get_table_count()
 
 
-class Sequence(object):
+class Sequence:
 
     """
     MAD-X sequence representation.
@@ -737,15 +743,13 @@ class Command(_MutableMapping):
         try:
             return self._attr[name]
         except KeyError:
-            pass
-        return _Mapping.__getattr__(self, name)
+            return _Mapping.__getattr__(self, name)
 
     def __getitem__(self, name):
         try:
             return self._data[name.lower()].value
         except KeyError:
-            pass
-        raise KeyError("Unknown command: {!r}".format(name))
+            raise KeyError("Unknown command: {!r}".format(name)) from None
 
     def __delitem__(self, name):
         raise NotImplementedError()
@@ -836,6 +840,21 @@ class ArrayAttribute(abc.Sequence):
         self._values[index] = value
         self._element[self._name] = self._values
 
+    def __eq__(self, other):
+        return self._values == other
+
+    def __lt__(self, other):
+        return self._values < other
+
+    def __le__(self, other):
+        return self._values <= other
+
+    def __gt__(self, other):
+        return self._values > other
+
+    def __ge__(self, other):
+        return self._values >= other
+
     def __iter__(self):
         return iter(self._values)
 
@@ -849,7 +868,7 @@ class ArrayAttribute(abc.Sequence):
         return str(self._values)
 
 
-class BaseElementList(object):
+class BaseElementList:
 
     """
     Immutable list of beam line elements.
@@ -872,7 +891,8 @@ class BaseElementList(object):
             try:
                 index = self.index(index)
             except ValueError:
-                raise KeyError("Unknown element: {!r}".format(index))
+                raise KeyError(
+                    "Unknown element: {!r}".format(index)) from None
         # _get_element accepts indices in the range [0, len-1]. The following
         # extends the accepted range to [-len, len+1], just like for lists:
         _len = len(self)
@@ -883,7 +903,10 @@ class BaseElementList(object):
             index += _len
         data = self._get_element(index)
         data['index'] = index
-        return Element(self._madx, data)
+        if data['base_type'] == 'sequence':
+            return Sequence(data['name'], self._madx)
+        else:
+            return Element(self._madx, data)
 
     def __len__(self):
         """Get number of elements."""
@@ -1006,10 +1029,9 @@ class CommandMap(_Mapping):
         try:
             data = madx._libmadx.get_defined_command(name)
         except ValueError:
-            pass
-        else:
-            return Command(madx, data)
-        raise KeyError("Not a MAD-X command name: {!r}".format(name))
+            raise KeyError(
+                "Not a MAD-X command name: {!r}".format(name)) from None
+        return Command(madx, data)
 
     def __contains__(self, name):
         return name.lower() in self._names
@@ -1041,14 +1063,35 @@ class Table(_Mapping):
     Loads individual columns from the MAD-X process lazily only on demand.
     """
 
-    def __init__(self, name, libmadx, _check=True):
+    def __init__(self, name, libmadx, *, columns='all', rows='all', _check=True):
         """Just store the table name for now."""
         self._name = name = name.lower()
         self._libmadx = libmadx
+        self._columns = columns
+        self._rows = rows
         self._cache = {}
         if _check and not libmadx.table_exists(name):
             raise ValueError("Invalid table: {!r}".format(name))
 
+    def selection(self, columns='selected', rows=None) -> "Table":
+        """
+        Return a Table object that only retrieves the specified rows and
+        columns by default.
+
+        :param columns: list of names or 'all' or 'selected'
+        :param rows: list of indices or 'all' or 'selected'
+
+        If only ``columns`` is given, ``rows`` defaults to 'selected'
+        unless ``columns`` is set to 'all' in which case it also defaults
+        to 'all'.
+        """
+        if rows is None:
+            rows = columns if isinstance(columns, str) else 'selected'
+        return Table(
+            self._name, self._libmadx,
+            columns=columns, rows=rows,
+            _check=False)
+
     def __getitem__(self, column):
         """Get the column data."""
         if isinstance(column, int):
@@ -1058,21 +1101,13 @@ class Table(_Mapping):
         except KeyError:
             return self.reload(column)
 
-    def _query(self, column):
-        """Retrieve the column data."""
-        try:
-            return self._libmadx.get_table_column(self._name, column.lower())
-        except ValueError:
-            pass
-        raise KeyError("Unknown table column: {!r}".format(column))
-
     def __iter__(self):
         """Iterate over all column names."""
         return iter(self.col_names())
 
     def __len__(self):
         """Return number of columns."""
-        return self._libmadx.get_table_column_count(self._name)
+        return self._libmadx.get_table_column_count(self._name, self._columns)
 
     def __repr__(self):
         return "<{} {!r}: {{{}}}>".format(
@@ -1093,13 +1128,27 @@ class Table(_Mapping):
         empty)."""
         return self._libmadx.get_table_selected_rows(self._name)
 
-    def col_names(self):
+    def col_names(self, columns=None):
         """Get list of all columns in the table."""
-        return self._libmadx.get_table_column_names(self._name)
+        if columns is None:
+            columns = self._columns
+        if isinstance(columns, str):
+            return self._libmadx.get_table_column_names(
+                self._name, selected=columns == 'selected')
+        else:
+            return columns
 
-    def row_names(self):
-        """Get table row names."""
-        return self._libmadx.get_table_row_names(self._name)
+    def row_names(self, rows=None):
+        """
+        Get table row names.
+
+        WARNING: using ``row_names`` after calling ``USE`` (before recomputing
+        the table) is unsafe and may lead to segmentation faults or incorrect
+        results.
+        """
+        if rows is None:
+            rows = self._rows
+        return self._libmadx.get_table_row_names(self._name, rows)
 
     @property
     def range(self):
@@ -1110,30 +1159,68 @@ class Table(_Mapping):
 
     def reload(self, column):
         """Reload (recache) one column from MAD-X."""
-        self._cache[column.lower()] = data = self._query(column)
+        try:
+            data = self._cache[column.lower()] = self.column(column)
+        except ValueError:
+            raise KeyError(
+                "Unknown table column: {!r}".format(column)) from None
         return data
 
-    def row(self, index, columns='all'):
+    def column(self, column: str, rows=None) -> np.ndarray:
+        """Retrieve all specified rows in the given column of the table.
+
+        :param column: column name
+        :param rows: a list of row indices or ``'all'`` or ``'selected'``
+        """
+        if rows is None:
+            rows = self._rows
+        return self._libmadx.get_table_column(self._name, column.lower(), rows)
+
+    def row(self, index, columns=None):
         """Retrieve one row from the table."""
+        if columns is None:
+            columns = self._columns
         return AttrDict(self._libmadx.get_table_row(self._name, index, columns))
 
-    def copy(self, columns=None):
+    def copy(self, columns=None, rows=None) -> dict:
         """
         Return a frozen table with the desired columns.
 
         :param list columns: column names or ``None`` for all columns.
         :returns: column data
-        :rtype: dict
         :raises ValueError: if the table name is invalid
         """
-        if columns is None:
-            columns = self
-        return {column: self[column] for column in columns}
+        if rows is None:
+            rows = columns if isinstance(columns, str) else self._rows
+        if rows == self._rows:
+            table = self
+        else:
+            table = self.select(rows=rows)
+        return {column: table[column] for column in self.col_names(columns)}
 
-    def dframe(self, columns=None):
-        """Return table as ``pandas.DataFrame``."""
+    def dframe(self, columns=None, rows=None, *, index=None):
+        """
+        Return table as ``pandas.DataFrame``.
+
+        :param columns: column names or 'all' or 'selected'
+        :param rows: row indices or 'all' or 'selected'
+        :param str index: column name or sequence to be used as index, or
+                          ``None`` for using the ``row_names``
+        :returns: column data as ``pandas.DataFrame``
+        :raises ValueError: if the table name is invalid
+
+        WARNING: using ``index=None`` is unsafe after calling ``USE``.
+        In this case, please manually specify another column to be used,
+        e.g. ``index="name"``.
+        """
         import pandas as pd
-        return pd.DataFrame(self.copy(columns), index=self.row_names())
+        if index is None:
+            index = self.row_names(rows)
+        elif isinstance(index, str):
+            index = util.remove_count_suffix_from_name(self[index])
+        else:
+            index = index
+        return pd.DataFrame(self.copy(columns, rows), index=index)
 
     def getmat(self, name, idx, *dim):
         s = () if isinstance(idx, int) else (-1,)
@@ -1231,7 +1318,7 @@ class VarParamList(_Mapping):
         return self._libmadx.num_globals()
 
 
-class Metadata(object):
+class Metadata:
 
     """MAD-X metadata (license info, etc)."""
 
