@@ -5,11 +5,16 @@ import re
 import os
 import sys
 import tempfile
+from collections import namedtuple
 from contextlib import contextmanager
+from enum import Enum
 from numbers import Number
 import collections.abc as abc
 
-from .types import (
+import numpy as np
+
+from cpymad.parsing import Parser
+from cpymad.types import (
     Range, Constraint,
     PARAM_TYPE_LOGICAL, PARAM_TYPE_INTEGER,
     PARAM_TYPE_DOUBLE, PARAM_TYPE_STRING, PARAM_TYPE_CONSTRAINT,
@@ -37,7 +42,7 @@ __all__ = [
 ordered_keys = dict.keys if sys.version_info >= (3, 6) else sorted
 
 
-def mad_quote(value):
+def mad_quote(value: str) -> str:
     """Add quotes to a string value."""
     if '"' not in value:
         return '"' + value + '"'
@@ -48,7 +53,7 @@ def mad_quote(value):
                      .format(value))
 
 
-def _fix_name(name):
+def _fix_name(name: str) -> str:
     if name.startswith('_'):
         raise AttributeError("Unknown item: {!r}! Did you mean {!r}?"
                              .format(name, name.strip('_') + '_'))
@@ -65,12 +70,12 @@ _re_element_internal = re_compile(r'^([a-z_][a-z0-9_.$]*)(:\d+)?$')
 _re_element_external = re_compile(r'^([a-z_][a-z0-9_.$]*)(\[\d+\])?$')
 
 
-def is_identifier(name):
+def is_identifier(name: str) -> bool:
     """Check if ``name`` is a valid identifier in MAD-X."""
     return bool(_re_is_identifier.match(name))
 
 
-def expr_symbols(expr):
+def expr_symbols(expr: str) -> set:
     """
     Return all symbols names used in an expression.
 
@@ -80,7 +85,7 @@ def expr_symbols(expr):
     return {m[0] for m in _re_symbol.findall(expr)}
 
 
-def name_from_internal(element_name):
+def name_from_internal(element_name: str) -> str:
     """
     Convert element name from internal representation to user API. Example:
 
@@ -98,13 +103,13 @@ def name_from_internal(element_name):
         name, count = _re_element_internal.match(element_name).groups()
     except AttributeError:
         raise ValueError("Not a valid MAD-X element name: {!r}"
-                         .format(element_name))
+                         .format(element_name)) from None
     if count is None or count == ':1':
         return name
     return name + '[' + count[1:] + ']'
 
 
-def _parse_element_name(element_name):
+def _parse_element_name(element_name: str) -> tuple:
     """
     Parse element name from user API. Example:
 
@@ -119,13 +124,13 @@ def _parse_element_name(element_name):
         name, count = _re_element_external.match(element_name).groups()
     except AttributeError:
         raise ValueError("Not a valid MAD-X element name: {!r}"
-                         .format(element_name))
+                         .format(element_name)) from None
     if count is None:
         return name, None
     return name, int(count[1:-1])
 
 
-def name_to_internal(element_name):
+def name_to_internal(element_name: str) -> str:
     """
     Convert element name from user API to internal representation. Example:
 
@@ -140,7 +145,7 @@ def name_to_internal(element_name):
     return name + ':' + str(1 if count is None else count)
 
 
-def normalize_range_name(name, elems=None):
+def normalize_range_name(name: str, elems=None) -> str:
     """Make element name usable as argument to the RANGE attribute."""
     if isinstance(name, tuple):
         return tuple(map(normalize_range_name, name))
@@ -160,7 +165,7 @@ QUOTED_PARAMS = {'file', 'halofile', 'sectorfile', 'trueprofile'
                  'echo', 'title', 'text', 'format', 'dir'}
 
 
-def format_param(key, value):
+def format_param(key: str, value) -> str:
     """
     Format a single MAD-X command parameter.
 
@@ -205,7 +210,7 @@ def format_param(key, value):
         return key + '=' + str(value)
 
 
-def _format_range(value):
+def _format_range(value) -> str:
     if isinstance(value, str):
         return normalize_range_name(value)
     elif isinstance(value, Range):
@@ -216,9 +221,13 @@ def _format_range(value):
     return begin + '/' + end
 
 
-def format_cmdpar(cmd, key, value):
+def format_cmdpar(cmd, key: str, value) -> str:
     """
     Format a single MAD-X command parameter.
+
+    :param cmd: A MAD-X Command instance for which an argument is to be formatted
+    :param key: name of the parameter
+    :param value: argument value
     """
     key = _fix_name(str(key).lower())
     cmdpar = cmd.cmdpar[key]
@@ -295,7 +304,7 @@ def format_cmdpar(cmd, key, value):
                     .format(key, value, type(value)))
 
 
-def format_command(*args, **kwargs):
+def format_command(*args, **kwargs) -> str:
     """
     Create a MAD-X command from its name and parameter list.
 
@@ -303,7 +312,6 @@ def format_command(*args, **kwargs):
     :param args: initial bareword command arguments (including command name!)
     :param kwargs: following named command arguments
     :returns: command string
-    :rtype: str
 
     Examples:
 
@@ -330,7 +338,74 @@ def format_command(*args, **kwargs):
 
 # validation of MAD-X expressions
 
-def _regex(expr):
+class T(Enum):
+    """Terminal/token type."""
+    WHITESPACE = 0
+    LPAREN     = 1
+    RPAREN     = 2
+    COMMA      = 3
+    SIGN       = 4
+    OPERATOR   = 5
+    SYMBOL     = 6
+    NUMBER     = 7
+    END        = 8
+
+    __str__ = __repr__ = lambda self: self.name
+
+
+class N(Enum):
+    """Nonterminal symbol."""
+    start            = 0
+    expression       = 1
+    expression_inner = 2
+    expression_tail  = 3
+    symbol_tail      = 4
+    argument_list    = 5
+    argument_tail    = 6
+
+    __str__ = __repr__ = lambda self: self.name
+
+
+grammar = {
+    N.start: [
+        [N.expression, T.END],
+    ],
+    N.expression: [
+        [T.WHITESPACE, N.expression],
+        [N.expression_inner],
+    ],
+    N.expression_inner: [
+        [T.SIGN, N.expression],
+        [T.LPAREN, N.expression, T.RPAREN, N.expression_tail],
+        [T.NUMBER, N.expression_tail],
+        [T.SYMBOL, N.symbol_tail],
+    ],
+    N.expression_tail: [
+        [T.WHITESPACE, N.expression_tail],
+        [T.SIGN, N.expression],
+        [T.OPERATOR, N.expression],
+        [],
+    ],
+    N.symbol_tail: [
+        [T.WHITESPACE, N.symbol_tail],
+        [T.LPAREN, N.argument_list, T.RPAREN, N.expression_tail],
+        [T.SIGN, N.expression],
+        [T.OPERATOR, N.expression],
+        [],
+    ],
+    N.argument_list: [
+        [T.WHITESPACE, N.argument_list],
+        [N.expression_inner, N.argument_tail],
+        [],
+    ],
+    N.argument_tail: [
+        [T.COMMA, N.argument_list],
+        [],
+    ],
+}
+
+
+def _regex(expr: str) -> callable:
     regex = re.compile(expr)
     def match(text, i):
         m = regex.match(text[i:])
@@ -338,30 +413,44 @@ def _regex(expr):
     return match
 
 
-def _choice(choices):
+def _choice(choices: str) -> callable:
     def match(text, i):
         return 1 if text[i] in choices else 0
     return match
 
 
-_expr_tokens = [
-    ('WHITESPACE',  _choice(' \t')),
-    ('LPAREN',      _choice('(')),
-    ('RPAREN',      _choice(')')),
-    ('OPERATOR',    _choice('+-/*^')),
-    ('SYMBOL',      _regex(r'[a-zA-Z_][a-zA-Z0-9_.]*(->[a-zA-Z_][a-zA-Z0-9_]*)?')),
-    ('NUMBER',      _regex(r'(\d+(\.\d*)?|\.\d+)([eE][+\-]?\d+)?')),
-]
+_expr_tokens = {
+    T.WHITESPACE:   _choice(' \t'),
+    T.LPAREN:       _choice('('),
+    T.RPAREN:       _choice(')'),
+    T.COMMA:        _choice(','),
+    T.SIGN:         _choice('+-'),
+    T.OPERATOR:     _choice('/*^'),
+    T.SYMBOL:       _regex(r'[a-zA-Z_][a-zA-Z0-9_.]*(->[a-zA-Z_][a-zA-Z0-9_]*)?'),
+    T.NUMBER:       _regex(r'(\d+(\.\d*)?|\.\d+)([eE][+\-]?\d+)?'),
+}
+
+_expr_parser = Parser(T, grammar, N.start)
 
 
-def _tokenize(tokens, expr):
+class Token(namedtuple('Token', ['type', 'start', 'length', 'expr'])):
+
+    @property
+    def text(self):
+        return self.expr[self.start:self.start + self.length]
+
+    def __repr__(self):
+        return '{}({!r})'.format(self.type, self.text)
+
+
+def tokenize(tokens, expr: str):
     i = 0
     stop = len(expr)
     while i < stop:
-        for tokname, tokmatch in tokens:
+        for toktype, tokmatch in tokens:
             l = tokmatch(expr, i)
             if l > 0:
-                yield tokname, i, l
+                yield Token(toktype, i, l, expr)
                 i += l
                 break
         else:
@@ -369,13 +458,12 @@ def _tokenize(tokens, expr):
                              .format(expr[i], expr[:i+1]))
 
 
-def check_expression(expr):
-
+def check_expression(expr: str):
     """
     Check if the given expression is a valid MAD-X expression that is safe to
     pass to :meth:`cpymad.madx.Madx.eval`.
 
-    :param str expr:
+    :param expr:
     :returns: True
     :raises ValueError: if the expression is ill-formed
 
@@ -383,60 +471,10 @@ def check_expression(expr):
     accepted by MAD-X and rejects valid but strange ones such as a number
     formatting '.' representing zero.
     """
-
-    expr = expr.strip()
-
-    def unexpected(tok, i, l):
-        return "Unexpected {} {!r} after {!r} in expression: {!r}".format(
-            tok, expr[i:i+l], expr[:i], expr)
-
-    _EXPRESSION = 'expression'
-    _OPERATOR = 'operator'
-    _ATOM = 'atom'
-    expect = _EXPRESSION
-    paren_level = 0
-
-    # expr =
-    #       symbol | number
-    #   |   '(' expr ')'
-    #   |   expr op expr
-
-    for tok, i, l in _tokenize(_expr_tokens, expr):
-        # ignore whitespace
-        if tok == 'WHITESPACE':
-            pass
-        # expr = symbol | number
-        elif tok in ('SYMBOL', 'NUMBER'):
-            if expect not in (_EXPRESSION, _ATOM):
-                raise ValueError(unexpected(tok, i, l))
-            expect = _OPERATOR
-        # expr = '(' expr ')'
-        elif tok == 'LPAREN':
-            if expect not in (_EXPRESSION, _ATOM):
-                raise ValueError(unexpected(tok, i, l))
-            paren_level += 1
-            expect = _EXPRESSION
-        elif tok == 'RPAREN':
-            if expect != _OPERATOR:
-                raise ValueError(unexpected(tok, i, l))
-            paren_level -= 1
-            expect = _OPERATOR
-        # expr = expr op expr
-        elif tok == 'OPERATOR':
-            if expect == _OPERATOR:
-                expect = _EXPRESSION
-            elif expect == _EXPRESSION and expr[i] in '+-':
-                expect = _ATOM
-            else:
-                raise ValueError(unexpected(tok, i, l))
-
-    if expect != _OPERATOR:
-        raise ValueError("Unexpected end-of-string in expression: {!r}"
-                         .format(expr))
-    if paren_level != 0:
-        raise ValueError("{} unclosed left-parens in expression: {!r}"
-                         .format(paren_level, expr))
-
+    expr = expr.strip().lower()
+    tokens = list(tokenize(list(_expr_tokens.items()), expr))
+    tokens.append(Token(T.END, len(expr), 0, expr))
+    _expr_parser.parse(tokens)  # raises ValueError
     return True
 
 
@@ -456,7 +494,7 @@ def temp_filename():
             pass
 
 
-class ChangeDirectory(object):
+class ChangeDirectory:
 
     """
     Context manager for temporarily changing current working directory.
@@ -489,3 +527,9 @@ class ChangeDirectory(object):
         """Exit 'with' context and restore old path."""
         if self._restore:
             self._chdir(self._restore)
+
+
+@np.vectorize
+def remove_count_suffix_from_name(name):
+    """Return the :N suffix from an element name."""
+    return name.rsplit(':', 1)[0]
